import fnmatch
import fractions
from collections import Counter
from collections import defaultdict
from pathlib import Path
from pathlib import PosixPath

import rich.text
import sage.all
import sage.matrix.constructor
import semver
import yaml
from sage.matrix import matrix_integer_dense
from sage.rings import integer
from sage.rings.integer_ring import ZZ

from autoeft import exceptions
from autoeft.combinatorics import young


def type_matches(op_type, selection):
    counter = op_type.to_counter()

    def _type_matches(select):
        for field_name, multiplicity in select.items():
            if isinstance(multiplicity, str):
                if multiplicity == "+":
                    if counter[field_name] == 0:
                        return False
                elif multiplicity.startswith(".."):
                    if counter[field_name] > int(multiplicity[2:]):
                        return False
                elif multiplicity.endswith(".."):
                    if counter[field_name] < int(multiplicity[:-2]):
                        return False
                elif r := multiplicity.find(".."):
                    if r < 0:
                        errmsg = f"Invalid range: '{multiplicity}'"
                        raise exceptions.InvalidRangeException(errmsg)
                    if counter[field_name] < int(multiplicity[:r]) or counter[
                        field_name
                    ] > int(multiplicity[(r + 2) :]):
                        return False
                else:
                    errmsg = f"Invalid range: '{multiplicity}'"
                    raise exceptions.InvalidRangeException(errmsg)
            elif isinstance(multiplicity, int):
                if counter[field_name] != multiplicity:
                    return False
            else:
                errmsg = f"Invalid range: '{multiplicity}'"
                raise exceptions.InvalidRangeException(errmsg)
        return True

    return any(_type_matches(select) for select in selection)


def type_matches_legacy(op_type, selection):
    return any(fnmatch.fnmatch(repr(op_type), pattern) for pattern in selection)


class Fraction(fractions.Fraction):
    def __rich__(self):
        return rich.text.Text(str(self))

    def __neg__(self):
        return Fraction(super().__neg__())

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v):
        return cls(v)

    @classmethod
    def encode(cls, v):
        if v.denominator == 1:
            return v.numerator
        return str(v)


class IntegerMatrix(matrix_integer_dense.Matrix_integer_dense):
    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v):
        if isinstance(v, matrix_integer_dense.Matrix_integer_dense):
            return v
        if isinstance(v, str):
            return sage.matrix.constructor.matrix(
                ZZ,
                [
                    [int(entry) for entry in row.lstrip("[").rstrip("]").split()]
                    for row in v.split("\n")
                ],
            )
        return ValueError


class Version(semver.Version):
    @classmethod
    def _parse(cls, version):
        return cls.parse(version)

    @classmethod
    def __get_validators__(cls):
        yield cls._parse


class Vector(Counter):
    def __add__(self, other):
        if isinstance(other, type(self)):
            (res := self.copy()).update(other)
            return res
        return NotImplemented

    def __iadd__(self, other):
        if isinstance(other, type(self)):
            self.update(other)
            return self
        return NotImplemented

    def __sub__(self, other):
        if isinstance(other, type(self)):
            (res := self.copy()).subtract(other)
            return res
        return NotImplemented

    def __isub__(self, other):
        if isinstance(other, type(self)):
            self.subtract(other)
            return self
        return NotImplemented

    def __mul__(self, other):
        if isinstance(other, (int, fractions.Fraction)):
            return type(self)({tab: coeff * other for tab, coeff in self.items()})
        return NotImplemented

    def __rmul__(self, other):
        if isinstance(other, (int, fractions.Fraction)):
            return type(self)({tab: other * coeff for tab, coeff in self.items()})
        return NotImplemented

    def __imul__(self, other):
        if isinstance(other, (int, fractions.Fraction)):
            res = {tab: coeff * other for tab, coeff in self.items()}
            self.clear()
            self.update(res)
            return self
        return NotImplemented

    def __truediv__(self, other):
        if isinstance(other, (int, fractions.Fraction)):
            return type(self)({tab: coeff / other for tab, coeff in self.items()})
        return NotImplemented

    def __itruediv__(self, other):
        if isinstance(other, (int, fractions.Fraction)):
            res = {tab: coeff / other for tab, coeff in self.items()}
            self.clear()
            self.update(res)
            return self
        return NotImplemented

    def __floordiv__(self, other):
        if isinstance(other, (int, fractions.Fraction)):
            return type(self)({tab: coeff // other for tab, coeff in self.items()})
        return NotImplemented

    def __ifloordiv__(self, other):
        if isinstance(other, (int, fractions.Fraction)):
            res = {tab: coeff // other for tab, coeff in self.items()}
            self.clear()
            self.update(res)
            return self
        return NotImplemented

    def __neg__(self):
        return -1 * self

    def __pos__(self):
        return 1 * self

    def __abs__(self):
        return type(self)({tab: abs(coeff) for tab, coeff in self.items()})

    def reduce(self):
        res = {tab: coeff for tab, coeff in self.items() if coeff}
        self.clear()
        self.update(res)
        return self

    def is_zero(self):
        return not bool(self.reduce())


class FlowSeq(list):
    pass


class FlowMap(dict):
    pass


class BlockMap(dict):
    pass


class YamlDumper(yaml.SafeDumper):
    def ignore_aliases(self, data):
        return True


def flowseq_representer(dumper, data):
    return dumper.represent_sequence("tag:yaml.org,2002:seq", data, flow_style=True)


def flowmap_representer(dumper, data):
    return dumper.represent_mapping("tag:yaml.org,2002:map", data, flow_style=True)


def blockmap_representer(dumper, data):
    return dumper.represent_mapping("tag:yaml.org,2002:map", data, flow_style=False)


def matrix_representer(dumper, data):
    return dumper.represent_scalar("tag:yaml.org,2002:str", str(data), style="|")


def version_representer(dumper, data):
    return dumper.represent_scalar("tag:yaml.org,2002:str", str(data))


def path_representer(dumper, data):
    return dumper.represent_scalar("tag:yaml.org,2002:str", str(data))


YamlDumper.add_representer(defaultdict, yaml.representer.Representer.represent_dict)
YamlDumper.add_representer(FlowSeq, flowseq_representer)
YamlDumper.add_representer(FlowMap, flowmap_representer)
YamlDumper.add_representer(BlockMap, blockmap_representer)
YamlDumper.add_representer(integer.Integer, yaml.representer.Representer.represent_int)
YamlDumper.add_representer(young.Partition, yaml.representer.Representer.represent_list)
YamlDumper.add_representer(
    matrix_integer_dense.Matrix_integer_dense,
    matrix_representer,
)
YamlDumper.add_representer(semver.Version, version_representer)
YamlDumper.add_representer(Version, version_representer)
YamlDumper.add_representer(Path, path_representer)
YamlDumper.add_representer(PosixPath, path_representer)


def non_redundant_json(*args, **kwargs):
    import json

    data = args[0]
    for datum in data["fields"].values():
        del datum["symmetries"]
    return json.dumps(data, *args[1:], **kwargs)


def esc(s: str) -> str:
    return s.replace("+", "_hc")



import json
import re
from typing import ClassVar
from typing import Dict
from typing import Optional

import pydantic
import rich.table
import rich.text

import autoeft
import autoeft.base.groups as symmetry_groups
import autoeft.base.representations as reprs
from autoeft import utils
from autoeft.tex import tex


class Symmetries(pydantic.BaseModel, frozen=True, extra=pydantic.Extra.forbid):
    """A collection of the Lorentz, SU(N) and U(1) symmetries."""

    lorentz_group: symmetry_groups.LorentzGroup = symmetry_groups.LorentzGroup()
    sun_groups: Dict[str, symmetry_groups.SUNGroup] = {}
    u1_groups: Dict[str, symmetry_groups.U1Group] = {}

    def __hash__(self):
        return hash(
            (self.lorentz_group, frozenset(self.sun_groups), frozenset(self.u1_groups)),
        )

    def __eq__(self, other):
        if not isinstance(other, Symmetries):
            return False
        lor = self.lorentz_group == other.lorentz_group
        sun = self.sun_groups == other.sun_groups
        u1 = self.u1_groups == other.u1_groups
        order = tuple(self.sun_groups.keys()) == tuple(other.sun_groups.keys())
        return lor and sun and u1 and order

    def __getitem__(self, key):
        all_groups = {
            self.lorentz_group.name: self.lorentz_group,
            **self.sun_groups,
            **self.u1_groups,
        }
        return all_groups[key]

    def __iter__(self):
        all_groups = {
            self.lorentz_group.name: self.lorentz_group,
            **self.sun_groups,
            **self.u1_groups,
        }
        return iter(all_groups)

    def __tex__(self):
        return r" \otimes ".join(
            map(tex, (self.lorentz_group, *self.sun_groups.values())),
        )

    @pydantic.root_validator
    def check_ambiguity(cls, values):
        if {"lorentz_group", "sun_groups", "u1_groups"} <= values.keys():
            names = [values["lorentz_group"].name]
            names += list(values["sun_groups"])
            names += list(values["u1_groups"])
            if len(names) != len(set(names)):
                errmsg = (
                    f"Duplicate group names are not allowed: '{names}'."
                    "\n"
                    "Please make sure each group name is unique."
                )
                raise ValueError(errmsg)
        return values

    @pydantic.validator("sun_groups", pre=True)
    def parse_sun_groups(cls, v):
        return {
            name: symmetry_groups.SUNGroup(name=name, **data)
            for name, data in v.items()
        }

    @pydantic.validator("u1_groups", pre=True)
    def parse_u1_groups(cls, v):
        return {
            name: symmetry_groups.U1Group(name=name, **data) for name, data in v.items()
        }


FieldName = pydantic.constr(
    regex=(
        rf"^[a-zA-Z][a-zA-Z0-9]*[a-zA-Z0-9{re.escape(autoeft.conjugation_symbol)}]$"
        r"|^[a-zA-Z]$"
    ),
)


class Field(pydantic.BaseModel, frozen=True, extra=pydantic.Extra.forbid):
    """Represent a field by its quantum numbers and representations.

    This class holds all information of a field provided in the model file.
    """

    name: FieldName
    generations: pydantic.PositiveInt = 1
    symmetries: Symmetries
    representations: Dict[str, reprs.Repr]
    conjugate: Optional[bool] = None
    anticommute: Optional[bool] = None
    tex: Optional[str] = None
    tex_hc: Optional[str] = None

    def __str__(self):
        return self.name

    def __tex__(self):
        return self.tex or rf"\mathtt{{{self.name}}}"

    def __rich__(self):
        return rich.text.Text(str(self), "bold")

    def __hash__(self):
        return hash(self.name)

    @pydantic.validator("name")
    def parse_name(cls, v):
        if v == "D":
            errmsg = f"{v} is a reserved name"
            raise ValueError(errmsg)
        return v

    @pydantic.validator("tex", always=True)
    def parse_tex(cls, v, values):
        if "name" in values:
            name = values["name"]
            return v or rf"\mathtt{{{name}}}"
        return v

    @pydantic.validator("tex_hc", always=True)
    def parse_tex_hc(cls, v, values):
        if {"tex", "conjugate"} <= values.keys():
            tex = values["tex"]
            return v or rf"{{{tex}}}^\dagger" if values["conjugate"] else tex
        return v

    @pydantic.validator("representations", pre=True)
    def parse_representations(cls, v, values):
        representations = {}
        if "symmetries" in values:
            lorentz_group = values["symmetries"].lorentz_group
            representations[lorentz_group.name] = reprs.LorentzRepr(
                helicity=v.get(lorentz_group.name, 0),
            )
            representations.update(
                {
                    name: reprs.SUNRepr(N=group.N, partition=v.get(name, []))
                    for name, group in values["symmetries"].sun_groups.items()
                },
            )
            representations.update(
                {
                    name: reprs.U1Repr(charge=v.get(name, 0))
                    for name in values["symmetries"].u1_groups
                },
            )
        return representations

    @pydantic.validator("conjugate", always=True)
    def parse_conjugate(cls, v, values):
        if v is not None:
            return v
        if {"symmetries", "representations"} <= values.keys():
            lorentz_group = values["symmetries"].lorentz_group
            sun_groups = values["symmetries"].sun_groups
            u1_groups = values["symmetries"].u1_groups
            if values["representations"][lorentz_group.name].is_complex():
                return True
            pseudo_reals = 0
            for group_name in sun_groups:
                if values["representations"][group_name].is_complex():
                    return True
                if values["representations"][group_name].is_pseudo_real():
                    pseudo_reals += 1
            if pseudo_reals % 2:
                return True
            for group_name in u1_groups:
                if values["representations"][group_name].is_complex():
                    return True
            return False
        return v

    @pydantic.validator("anticommute", always=True)
    def parse_anticommute(cls, v, values):
        if {"symmetries", "representations"} <= values.keys():
            lorentz_group = values["symmetries"].lorentz_group
            helicity = values["representations"][lorentz_group.name].helicity
            return v if v is not None else helicity.denominator == 2
        return v

    @property
    def helicity(self) -> utils.Fraction:
        return self.representations[self.symmetries.lorentz_group.name].helicity

    def is_fermionic(self) -> bool:
        return self.helicity.denominator == 2

    def conjugated_name(self, conjugation_symbol: str = autoeft.conjugation_symbol):
        if self.name.endswith(conjugation_symbol):
            return self.name.rstrip(conjugation_symbol)
        return self.name + conjugation_symbol

    def hermitian_conjugated(self) -> "Field":
        """Return the hermitian conjugated field."""
        data = {}
        data["name"] = self.conjugated_name()
        data["tex"], data["tex_hc"] = self.tex_hc, self.tex
        data["representations"] = {
            group: sym_repr.conjugated()
            for group, sym_repr in self.representations.items()
        }
        return self.copy(update=data)


class Model(pydantic.BaseModel):
    """Represents a model by its symmetries and fields.

    This class holds all information provided in the model file.
    """

    name: str
    description: Optional[str] = None
    symmetries: Symmetries
    fields: Dict[str, Field]

    def __str__(self):
        if self.description:
            return f"{self.name}: {self.description}"
        return self.name

    def __tex__(self):
        def ___tex__():
            lorentz_group = self.symmetries.lorentz_group
            sun_groups = self.symmetries.sun_groups
            u1_groups = self.symmetries.u1_groups
            c = "(+)"
            ac = "(-)"
            alignment = "lr" + "r" * (len(sun_groups) + len(u1_groups)) + "rc"
            yield rf"\begin{{longtable}}{{{alignment}}}"
            yield r"\toprule"
            top_row = f"Field & ${tex(lorentz_group)}$ & "
            if sun_groups:
                top_row += " & ".join(f"${tex(g)}$" for g in sun_groups.values())
                top_row += " & "
            if u1_groups:
                top_row += " & ".join(f"${tex(g)}$" for g in u1_groups.values())
                top_row += " & "
            top_row += r"$g$ & $\pm$ \\"
            yield top_row
            yield r"\midrule"
            for f in self.fields.values():
                field_row = f"${tex(f)}$ & ${f.representations[lorentz_group.name]}$ & "
                if sun_groups:
                    field_row += " & ".join(
                        f"${tex(f.representations[g])}$" for g in sun_groups
                    )
                    field_row += " & "
                if u1_groups:
                    field_row += " & ".join(
                        f"${tex(f.representations[g])}$" for g in u1_groups
                    )
                    field_row += " & "
                field_row += (
                    rf"${f.generations}$ & ${c if not f.anticommute else ac}$ \\"
                )
                yield field_row
            yield r"\bottomrule"
            yield r"\end{longtable}"

        return "\n".join(___tex__())

    def __rich__(self):
        lorentz_group = self.symmetries.lorentz_group
        sun_groups = self.symmetries.sun_groups
        u1_groups = self.symmetries.u1_groups
        table = rich.table.Table(
            title=self.name,
            caption=self.description,
            show_edge=False,
            row_styles=["", "dim"],
        )
        table.add_column("Field")
        table.add_column(lorentz_group, justify="right")
        for sun_group in sun_groups:
            table.add_column(sun_group, justify="right")
        for u1_group in u1_groups:
            table.add_column(u1_group, justify="right")
        table.add_column("g", justify="right")
        table.add_column("+/-", justify="center")
        for f in self.fields.values():
            table.add_row(
                f,
                f.representations[lorentz_group.name],
                *(f.representations[g] for g in sun_groups),
                *(f.representations[g] for g in u1_groups),
                str(f.generations),
                "(+)" if not f.anticommute else "(-)",
            )
        return table

    def __eq__(self, other):
        if not isinstance(other, Model):
            return False
        return self.symmetries == other.symmetries and self.dict() == other.dict()

    @pydantic.validator("fields", pre=True)
    def parse_fields(cls, v, values):
        fields = {}
        if "symmetries" in values:
            for name, data in v.items():
                if name in fields:
                    errmsg = (
                        f"Found multiple definitions for '{name}'"
                        "\nConsider setting the 'conjugate' option to 'False'"
                    )
                    raise ValueError(errmsg)
                fields[name] = field = Field(
                    name=name,
                    symmetries=values["symmetries"],
                    **data,
                )
                if field.conjugate:
                    hc_field = field.hermitian_conjugated()
                    if hc_field.name in fields:
                        errmsg = (
                            f"Found multiple definitions for '{hc_field.name}'."
                            "\nConsider setting the 'conjugate' option to 'False'"
                        )
                        raise ValueError(errmsg)
                    fields[hc_field.name] = hc_field
        return fields

    @classmethod
    def parse_json(cls, data) -> "Model":
        model_data = json.loads(data)
        lorentz_group = symmetry_groups.LorentzGroup(
            **model_data["symmetries"]["lorentz_group"],
        )
        sun_groups = {}
        for name, group in model_data["symmetries"]["sun_groups"].items():
            sun_groups[name] = symmetry_groups.SUNGroup(**group)
        u1_groups = {}
        for name, group in model_data["symmetries"]["u1_groups"].items():
            u1_groups[name] = symmetry_groups.U1Group(**group)
        del model_data["symmetries"]
        symmetries = Symmetries.construct(
            lorentz_group=lorentz_group,
            sun_groups=sun_groups,
            u1_groups=u1_groups,
        )
        fields = {}
        for name, field in model_data["fields"].items():
            representations = {}
            for group, representation in field["representations"].items():
                if group == symmetries.lorentz_group.name:
                    representations[group] = representation["helicity"]
                elif group in symmetries.sun_groups:
                    representations[group] = representation["partition"]
                elif group in symmetries.u1_groups:
                    representations[group] = representation["charge"]
            del field["representations"]
            fields[name] = Field(
                **field,
                symmetries=symmetries,
                representations=representations,
            )
        del model_data["fields"]
        return cls.construct(**model_data, symmetries=symmetries, fields=fields)

    class Config:
        extra = pydantic.Extra.forbid
        json_dumps = utils.non_redundant_json
        json_encoders: ClassVar[dict] = {utils.Fraction: utils.Fraction.encode}


def main(args=None):
    import importlib.resources
    import sys

    import autoeft

    try:
        if sys.version_info >= (3, 9):
            sample = (
                importlib.resources.files(autoeft.__program__)
                .joinpath("sm.yml")
                .read_text()
            )
        else:
            sample = importlib.resources.read_text(autoeft.__program__, "sm.yml")
    except FileNotFoundError:
        sample = "name: <sample-model>\nsymmetries: {}\nfields: {}\n"
    sys.stdout.write(sample)
    return 0



import itertools
from collections import Counter
from collections import defaultdict
from math import prod
from typing import Counter as Count
from typing import Dict
from typing import Iterable
from typing import List
from typing import Optional
from typing import Tuple

import sage.all
import sage.combinat.permutation
import sage.combinat.sf.sf
import sage.combinat.skew_tableau
import sage.combinat.symmetric_group_algebra
import sage.groups.perm_gps.permgroup_named
import sage.matrix.constructor
from sage.libs.lrcalc import lrcalc
from sage.rings.rational_field import QQ

from autoeft import utils
from autoeft.combinatorics import generators
from autoeft.combinatorics import young


def lr_rule(
    lhs: young.Partition,
    rhs: young.Partition,
    N: Optional[int],
    factor: int = 1,
) -> Count[young.Partition]:
    """Interface to SageMath's Littlewood-Richardson rule."""
    return +Counter(
        {
            young.Partition(p): int(c * factor)
            for p, c in lrcalc.mult(lhs, rhs, N).items()
        },
    )


def multi_lr_rule(
    partitions: Iterable[young.Partition],
    N: Optional[int],
    initials: Optional[Count[young.Partition]] = None,
) -> Count[young.Partition]:
    it = iter(partitions)
    try:
        rhs = next(it)
    except StopIteration:
        return Counter(initials)
    if initials is None:
        initials = Counter({young.Partition([]): 1})
    result = Counter()
    for lhs, coeff in initials.items():
        result.update(lr_rule(lhs, rhs, N, coeff))
    return +multi_lr_rule(it, N, result)


def lr_skew_tableaux(
    outer: young.Partition,
    inner: young.Partition,
    weight: young.Partition,
) -> List[young.Tableau]:
    """Return all Littlewood-Richardson tableaux.

    A Littlewood-Richardson tableau is a semistandard skew tableau
    of shape outer/inner and the given weight
    which is also a valid lattice permutation.
    """
    SSSYTx = sage.combinat.skew_tableau.SemistandardSkewTableaux_shape_weight
    return sorted(
        young.Tableau(skew_tableau)
        for skew_tableau in SSSYTx([outer, inner], weight)
        if skew_tableau.to_word().is_yamanouchi()
    )


def plethysm(lhs: young.Partition, rhs: young.Partition) -> Count[young.Partition]:
    """Interface to SageMath's implementation of the plethysm.

    The order matches the definition of the plethysm operation given in
    arXiv:2005.00008 [hep-ph],
    i.e., lhs * rhs = rhs(lhs).
    """
    if not lhs:
        return Counter({young.Partition(): 1})
    schur = sage.combinat.sf.sf.SymmetricFunctions(QQ).s()
    return +Counter({young.Partition(p): c for p, c in schur[rhs].plethysm(schur[lhs])})


def direct_product_decomposition(
    lhs: young.Partition,
    rhs: young.Partition,
    n: int,
    factor: int = 1,
) -> Count[young.Partition]:
    return multi_direct_product_decomposition([lhs, rhs], n, factor)


def multi_direct_product_decomposition(
    partitions: Iterable[young.Partition],
    n: int,
    factor: int = 1,
) -> Count[young.Partition]:
    SG = sage.groups.perm_gps.permgroup_named.SymmetricGroup(n)
    order = SG.order()
    character_table = SG.character_table()
    representations = {
        young.Partition(cc.partition()): i
        for i, cc in enumerate(SG.conjugacy_classes_iterator())
    }
    products = [representations[p] for p in partitions]
    return +Counter(
        {
            representation: sum(
                cc.cardinality()
                * character_table[lam, i]
                * prod(character_table[r, i] for r in products)
                for i, cc in enumerate(SG.conjugacy_classes_iterator())
            )
            * factor
            / order
            for representation, lam in representations.items()
        },
    )


def CGC(
    N: int,
    irrepr: young.Partition,
    j: int,
    i: int,
    decomp: List[Tuple[young.Partition, int]],
    vector: Tuple[int, ...],
) -> utils.Fraction:
    return sum(
        representation_matrix(irrepr, p).inverse()[j, i]
        * prod(representation_matrix(mu, p)[x, y] for (mu, x), y in zip(decomp, vector))
        for p in sage.combinat.permutation.StandardPermutations_n(N)
    ) / sage.combinat.symmetric_group_algebra.kappa(irrepr)


def construct_CGCs(
    N: int,
    partitions: List[young.Partition],
    target: young.Partition,
    multiplicity: int,
) -> sage.matrix.constructor.Matrix:
    basis = list(
        itertools.product(*(range(part.dimension_SN()) for part in partitions)),
    )
    vectors = []
    for vector1 in basis:
        vectors.append(
            [
                CGC(N, target, 0, 0, list(zip(partitions, vector2)), vector1)
                for vector2 in basis
            ],
        )
        if (matrix := sage.matrix.constructor.matrix(vectors)).rank() == multiplicity:
            return matrix[matrix.pivot_rows(), :], basis
    errmsg = "Not enough linear independent vectors found"
    raise RuntimeError(errmsg)


def join_CGCs(
    cgcs: Iterable[Count[Tuple[int, ...]]],
    initials: Optional[Count[Tuple[Tuple[int, ...], ...]]] = None,
) -> Count[Tuple[Tuple[int, ...], ...]]:
    it = iter(cgcs)
    try:
        rhs = next(it)
    except StopIteration:
        return Counter(initials)
    if initials is None:
        initials = Counter({(): 1})
    result = Counter()
    for lhs, coeff1 in initials.items():
        for vector, coeff2 in rhs.items():
            if not lhs:
                merged = tuple((entry,) for entry in vector)
            else:
                merged = tuple((*a, b) for a, b in zip(lhs, vector))
            result[merged] = coeff1 * coeff2
    return join_CGCs(it, result)


# cache representation matrices
representation_matrices: Dict[
    young.Partition,
    Dict[sage.combinat.permutation.Permutation, sage.matrix.constructor.Matrix],
] = defaultdict(dict)


def representation_matrix(
    irrepr: young.Partition,
    group_element: sage.combinat.permutation.Permutation,
) -> sage.matrix.constructor.Matrix:
    try:
        return representation_matrices[irrepr][group_element]
    except KeyError:
        D_reflection, D_cycle, D_cycle_I = generators.get_generators(irrepr)
        dim = irrepr.dimension_SN()
        matrix = prod(
            (
                D_cycle ** (i - 1) * D_reflection * D_cycle_I ** (i - 1)
                for i in group_element.reduced_word()
            ),
            start=sage.matrix.constructor.matrix.identity(QQ, dim),
        )
        representation_matrices[irrepr][group_element] = matrix
        return matrix



import itertools
from typing import Iterable
from typing import Iterator
from typing import List
from typing import Optional
from typing import Tuple

import rich.text
import sage.all
import sage.combinat.partition
import sage.combinat.tableau

Composition = Tuple[int, ...]
Dynkin = List[int]


class Partition(tuple):
    """A partition of a non-negative integer n.

    Representet by a non-increasing tuple of positive integers with total sum n.
    """

    def __new__(cls, partition: Optional[Iterable[int]] = None):
        if partition is None:
            return super().__new__(cls)
        stripped_part = tuple(itertools.takewhile(lambda x: x > 0, partition))
        if not cls.is_partition(stripped_part):
            errmsg = f"{stripped_part} is not a valid partition"
            raise ValueError(errmsg)
        return super().__new__(cls, stripped_part)

    def __repr__(self):
        return str(self)

    def __str__(self):
        return f"[{','.join(str(i) for i in self)}]"

    def __rich__(self):
        return rich.text.Text.assemble(
            ("[", "bold"),
            ",".join(str(i) for i in self),
            ("]", "bold"),
        )

    def dimension_SN(self) -> int:
        SYTx = sage.combinat.tableau.StandardTableaux_shape
        return SYTx(
            sage.combinat.partition._Partitions(self),  # noqa: SLF001
        ).cardinality()

    def dimension_SUN(self, max_entry: Optional[int] = None) -> int:
        SSYTx = sage.combinat.tableau.SemistandardTableaux
        if max_entry:
            return SSYTx(shape=self, max_entry=max_entry).cardinality()
        return SSYTx(shape=self).cardinality()

    def normal_tableau(self) -> "Tableau":
        num = 0
        return Tableau((num := num + 1 for _ in range(row_len)) for row_len in self)

    def standard_tableaux(self) -> List["Tableau"]:
        SYTx = sage.combinat.tableau.StandardTableaux_shape
        return sorted(
            Tableau(t)
            for t in SYTx(sage.combinat.partition._Partitions(self))  # noqa: SLF001
        )

    def semistandard_tableaux(self, max_entry: Optional[int] = None) -> List["Tableau"]:
        SSYTx = sage.combinat.tableau.SemistandardTableaux
        if max_entry:
            return sorted(Tableau(t) for t in SSYTx(shape=self, max_entry=max_entry))
        return sorted(Tableau(t) for t in SSYTx(shape=self))

    def to_dynkin(self, N: int) -> Dynkin:
        """Return the Dynkin labels corresponding to this partition.

        For a partition (p1, p2, p3, ..., pn) the corresponding
        Dynkin labels are [p1 - p2, p2 - p3, ..., pn-1 - pn].
        """
        extended_part = list(self) + [0] * (N - len(self))
        return [p1 - p2 for p1, p2 in zip(extended_part, extended_part[1:])]

    @classmethod
    def from_dynkin(cls, dynkin: Dynkin) -> "Partition":
        """Construct partition from Dynkin labels.

        For the Dynkin labels [l1, l2, l3, ..., ln] the corresponding
        partition is given by (l1 + l2 + l3 + ... + ln, l2 + l3 + ... + ln, ..., ln).
        """
        return cls(sum(dynkin[label:]) for label in range(len(dynkin)))

    @classmethod
    def partitions(cls, m: int) -> Iterator["Partition"]:
        yield from (Partition(p) for p in sage.combinat.partition.Partitions_n(m))

    @staticmethod
    def is_partition(partition: Iterable[int], n: Optional[int] = None) -> bool:
        """Check if given partition is valid.

        If n is given also check if its a partition of n.
        Allow positive entries only.
        """
        if any(x <= 0 for x in partition):
            return False
        if list(partition) != sorted(partition, reverse=True):
            return False
        if n is not None:
            return sum(partition) == n
        return True

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v):
        return cls(v)


class Tableau(tuple):
    """A tableau of given shape and content.

    Represented by a tuple of tuples containing arbitrary elements.
    """

    def __new__(cls, tableau: Optional[Iterable[Iterable]] = None):
        if tableau is None:
            return super().__new__(cls)
        return super().__new__(cls, (tuple(row) for row in tableau if row))

    def __contains__(self, item):
        if isinstance(item, int):
            return any(item in row for row in self)
        return super().__contains__(item)

    def __getitem__(self, key):
        err_msg = (
            "tableau indices must be an integer,"
            " two integers, a slice, or two slices,"
            f" not {type(key).__name__}"
        )
        if isinstance(key, int):
            return list(super().__getitem__(key))
        if isinstance(key, slice):
            return Tableau(super().__getitem__(key))
        if isinstance(key, tuple):
            if len(key) > 2:
                raise TypeError(err_msg)
            if isinstance(key[0], int) and isinstance(key[1], int):
                return super().__getitem__(key[0])[key[1]]
            if isinstance(key[0], int) and isinstance(key[1], slice):
                return list(super().__getitem__(key[0])[key[1]])
            if isinstance(key[0], slice) and isinstance(key[1], slice):
                return Tableau(row[key[1]] for row in super().__getitem__(key[0]))
            if isinstance(key[0], slice) and isinstance(key[1], int):
                if key[1] >= max(self.shape):
                    errmsg = "tuple index out of range"
                    raise IndexError(errmsg)
                return [
                    row[key[1]]
                    for row in super().__getitem__(key[0])
                    if key[1] < len(row)
                ]
        raise TypeError(err_msg)

    @property
    def shape(self) -> Partition:
        """Return the shape of the tableau."""
        return Partition(len(row) for row in self)

    def is_ssyt(self) -> bool:
        return sage.combinat.tableau.Tableau(self).is_semistandard()

    def transposed(self) -> "Tableau":
        """Return the transposed tableau."""
        return Tableau(self[:, c] for c in range(max(self.shape, default=0)))

    def to_lists(self) -> List[list]:
        return [list(row) for row in self]

    def replaced(self, element, replacement) -> Iterator["Tableau"]:
        for i, row in enumerate(self):
            for j, entry in enumerate(row):
                if entry == element:
                    tableau = self.to_lists()
                    tableau[i][j] = replacement
                    yield Tableau(tableau)

    def flatten(self) -> tuple:
        return tuple(entry for row in self for entry in row)



import importlib.resources
import itertools
import re
import sys
from collections import defaultdict
from pathlib import Path
from typing import Dict
from typing import Iterator
from typing import Tuple

import numpy as np
import rich.progress
import rich.table
import sage.all
import sage.combinat.permutation
import sage.combinat.symmetric_group_algebra
import sage.matrix.constructor
import sage.modules.free_module_element
import scipy.sparse

import autoeft
from autoeft import exceptions
from autoeft.combinatorics import young

print = autoeft.print  # noqa: A001


default_module = __name__.rsplit(".", 1)[0]
default_path = Path().cwd() / "gens"
N_pattern = re.compile(r"^S_(\d+)$")
P_pattern = re.compile(r"^P_((?:\d+_)+)(cycle|reflection)\.npz$")


# cache young basis
young_basis: Dict[young.Partition, list] = {}
# cache generators
generators: Dict[
    young.Partition,
    Tuple[
        sage.matrix.constructor.Matrix,
        sage.matrix.constructor.Matrix,
        sage.matrix.constructor.Matrix,
    ],
] = {
    young.Partition(): (
        sage.matrix.constructor.Matrix.identity(1),
        sage.matrix.constructor.Matrix.identity(1),
        sage.matrix.constructor.Matrix.identity(1),
    ),
    young.Partition([1]): (
        sage.matrix.constructor.Matrix.identity(1),
        sage.matrix.constructor.Matrix.identity(1),
        sage.matrix.constructor.Matrix.identity(1),
    ),
}


def get_young_basis(irrepr: young.Partition) -> list:
    try:
        return young_basis[irrepr]
    except KeyError:
        tableaux = irrepr.standard_tableaux()
        normal_tableau = tableaux[0]
        young_operator = sage.combinat.symmetric_group_algebra.e_hat(normal_tableau)
        algebra = young_operator.parent()
        basis = [
            algebra.left_action_product(
                sage.combinat.symmetric_group_algebra.pi_ik(normal_tableau, tableau),
                young_operator,
            )
            for tableau in tableaux
        ]
        young_basis[irrepr] = basis
        return basis


def get_path_generators(generators_path: Path, reflection_path: Path, cycle_path: Path):
    D_reflection = sage.matrix.constructor.matrix(
        scipy.sparse.load_npz(generators_path / reflection_path).toarray(),
    )
    D_cycle = sage.matrix.constructor.matrix(
        scipy.sparse.load_npz(generators_path / cycle_path).toarray(),
    )
    return D_reflection, D_cycle


def get_resource_generators(
    generators_path: Path,
    reflection_path: Path,
    cycle_path: Path,
):
    with importlib.resources.files(default_module).joinpath(
        generators_path,
        reflection_path,
    ).open("rb") as reflection_file:
        D_reflection = sage.matrix.constructor.matrix(
            scipy.sparse.load_npz(reflection_file).toarray(),
        )
    with importlib.resources.files(default_module).joinpath(
        generators_path,
        cycle_path,
    ).open("rb") as cycle_file:
        D_cycle = sage.matrix.constructor.matrix(
            scipy.sparse.load_npz(cycle_file).toarray(),
        )
    return D_reflection, D_cycle


def get_resource_generators_lagacy(
    generators_path: str,
    reflection_path: Path,
    cycle_path: Path,
):
    with importlib.resources.open_binary(
        f"{default_module}.{generators_path}",
        reflection_path,
    ) as reflection_file:
        D_reflection = sage.matrix.constructor.matrix(
            scipy.sparse.load_npz(reflection_file).toarray(),
        )
    with importlib.resources.open_binary(
        f"{default_module}.{generators_path}",
        cycle_path,
    ) as cycle_file:
        D_cycle = sage.matrix.constructor.matrix(
            scipy.sparse.load_npz(cycle_file).toarray(),
        )
    return D_reflection, D_cycle


def get_generators(
    irrepr: young.Partition,
) -> Tuple[
    sage.matrix.constructor.Matrix,
    sage.matrix.constructor.Matrix,
    sage.matrix.constructor.Matrix,
]:
    try:
        return generators[irrepr]
    except KeyError:
        N = sum(irrepr)
        str_irrepr = "_".join(str(i) for i in irrepr)
        reflection_path = Path(f"P_{str_irrepr}_reflection.npz")
        cycle_path = Path(f"P_{str_irrepr}_cycle.npz")
        try:
            generators_path = default_path / f"S_{N}"
            D_reflection, D_cycle = get_path_generators(
                generators_path,
                reflection_path,
                cycle_path,
            )
        except FileNotFoundError:
            try:
                if sys.version_info >= (3, 9):
                    generators_path = Path("gens", f"S_{N}")
                    D_reflection, D_cycle = get_resource_generators(
                        generators_path,
                        reflection_path,
                        cycle_path,
                    )
                else:
                    generators_path = f"gens.S_{N}"
                    D_reflection, D_cycle = get_resource_generators_lagacy(
                        generators_path,
                        reflection_path,
                        cycle_path,
                    )
            except (ModuleNotFoundError, FileNotFoundError) as e:
                errmsg = (
                    "Required generator matrices for irreducible representation"
                    f" {irrepr} not found."
                    "\n"
                    "You can try to generate them using 'autoeft generators'."
                )
                raise exceptions.MissingResourcesError(errmsg) from e
        D_cycle_I = D_cycle.inverse()
        generators[irrepr] = D_reflection, D_cycle, D_cycle_I
        return D_reflection, D_cycle, D_cycle_I


def generator_matrices(
    irrepr: young.Partition,
) -> Tuple[scipy.sparse.csc_matrix, scipy.sparse.csc_matrix]:
    with rich.progress.Progress(
        rich.progress.TextColumn("{task.description}", style="bold"),
        rich.progress.BarColumn(),
        rich.progress.MofNCompleteColumn(),
        rich.progress.TimeRemainingColumn(compact=True, elapsed_when_finished=True),
        console=autoeft.console,
    ) as progress:
        N = sum(irrepr)
        task = progress.add_task(f"{irrepr!s:<{2*N+1}}", total=None)
        group = sage.combinat.permutation.StandardPermutations_n(N)
        reflection = group.simple_reflection(1)
        cycle = sage.combinat.permutation.from_cycles(N, [list(range(1, N + 1))])
        basis = get_young_basis(irrepr)
        dim = len(basis)
        algebra = basis[0].parent()
        A = sage.matrix.constructor.matrix(
            sage.modules.free_module_element.vector(e) for e in basis
        )
        D_reflection = scipy.sparse.dok_array((dim, dim), dtype=np.int32)
        D_cycle = scipy.sparse.dok_array((dim, dim), dtype=np.int32)
        progress.update(task, total=dim)
        for i, element in enumerate(basis):
            Y_reflection = sage.modules.free_module_element.vector(
                algebra.left_action_product(reflection, element),
            )
            X_reflection = A.solve_left(Y_reflection)
            for j, value in enumerate(X_reflection):
                if value:
                    D_reflection[j, i] = value
            Y_cycle = sage.modules.free_module_element.vector(
                algebra.left_action_product(cycle, element),
            )
            X_cycle = A.solve_left(Y_cycle)
            for j, value in enumerate(X_cycle):
                if value:
                    D_cycle[(j, i)] = value
            progress.advance(task, 1)
    return D_reflection.tocsc(), D_cycle.tocsc()


def save_matrices(
    matrices_path: Path,
    irrepr: young.Partition,
    overwrite: bool = False,
):
    str_irrepr = "_".join(str(i) for i in irrepr)
    reflection_path = matrices_path / f"P_{str_irrepr}_reflection.npz"
    cycle_path = matrices_path / f"P_{str_irrepr}_cycle.npz"
    N = sum(irrepr)
    if overwrite or not (reflection_path.is_file() and cycle_path.is_file()):
        D_reflection, D_cycle = generator_matrices(irrepr)
        scipy.sparse.save_npz(reflection_path, D_reflection)
        scipy.sparse.save_npz(cycle_path, D_cycle)
        print(f"[white b]{irrepr!s:<{2*N+1}}[/] [dim]calculated, saved[/]", dest="log")
    else:
        print(f"[white b]{irrepr!s:<{2*N+1}}[/] [dim]already exists, skipped[/]")


def collect_generators_path(matrices_path: Path) -> Iterator[tuple]:
    for group_path in matrices_path.glob("S_*/"):
        if match_N := N_pattern.match(group_path.name):
            N = int(match_N[1])
            for partition_path in group_path.glob("P_*.npz"):
                if match_P := P_pattern.match(partition_path.name):
                    partition = young.Partition(
                        int(i) for i in match_P[1].split("_") if i
                    )
                    assert sum(partition) == N
                    try:
                        printed_path = partition_path.relative_to(Path().cwd())
                    except ValueError:
                        printed_path = partition_path
                    yield N, partition, match_P[2], printed_path


def collect_generators_resource() -> Iterator[tuple]:
    resource = importlib.resources.files(default_module).joinpath("gens")
    for group_resource in resource.iterdir():
        if (
            match_N := N_pattern.match(group_resource.name)
        ) and group_resource.is_dir():
            N = int(match_N[1])
            for partition_resource in group_resource.iterdir():
                if (
                    match_P := P_pattern.match(partition_resource.name)
                ) and partition_resource.is_file():
                    partition = young.Partition(
                        int(i) for i in match_P[1].split("_") if i
                    )
                    assert sum(partition) == N
                    with importlib.resources.as_file(
                        partition_resource,
                    ) as partition_path:
                        yield N, partition, match_P[2], partition_path


def collect_generators_resource_lagacy() -> Iterator[tuple]:
    resource = f"{default_module}.gens"
    for group_name in importlib.resources.contents(resource):
        if match_N := N_pattern.match(group_name):
            N = int(match_N[1])
            for partition_name in importlib.resources.contents(resource + f".S_{N}"):
                if match_P := P_pattern.match(partition_name):
                    partition = young.Partition(
                        int(i) for i in match_P[1].split("_") if i
                    )
                    assert sum(partition) == N
                    with importlib.resources.path(
                        resource + f".S_{N}",
                        partition_name,
                    ) as partition_path:
                        yield N, partition, match_P[2], partition_path


def collect_generators(matrices_path: Path) -> dict:
    system_generators = defaultdict(lambda: defaultdict(lambda: defaultdict(dict)))
    for N, part, kind, path in collect_generators_path(matrices_path):
        system_generators[N][part][kind]["path"] = path
    if sys.version_info >= (3, 9):
        for N, part, kind, path in collect_generators_resource():
            system_generators[N][part][kind]["resource"] = path
    else:
        for N, part, kind, path in collect_generators_resource_lagacy():
            system_generators[N][part][kind]["resource"] = path
    return system_generators


def show_generators(matrices_path: Path):
    system_generators = collect_generators(matrices_path)
    caption = (
        f"This table shows all generators that {autoeft.program}"
        " loads with the given options."
    )
    table = rich.table.Table(title="Generators", caption=caption, highlight=True)
    table.add_column("N")
    table.add_column("Partition")
    table.add_column("Cycle", overflow="fold")
    table.add_column("Reflection", overflow="fold")
    for N in sorted(system_generators):
        element = itertools.chain([str(N)], itertools.repeat(None))
        for partition in sorted(system_generators[N], reverse=True):
            system_generator = system_generators[N][partition]
            print_cycle = print_reflection = "[bold red]MISSING[/bold red]"
            if "cycle" in system_generator:
                cycle = system_generator["cycle"].get(
                    "path",
                    system_generator["cycle"]["resource"],
                )
                print_cycle = f"[link=file:{cycle.absolute()}]{cycle}[/link]"
            if "reflection" in system_generator:
                reflection = system_generator["reflection"].get(
                    "path",
                    system_generator["reflection"]["resource"],
                )
                print_reflection = (
                    f"[link=file:{reflection.absolute()}]{reflection}[/link]"
                )
            table.add_row(next(element), str(partition), print_cycle, print_reflection)
        table.add_section()
    print(table)


def main(args) -> int:
    output_path = args.output.resolve()
    if args.N is not None:
        N = args.N
        if N < 2:
            print(f"Generators for [b]S{N}[/b] are not required.")
            return 1
        print(f"Constructing generators for [b]S{N}[/b]")
        output_path /= f"S_{N}"
        output_path.mkdir(parents=True, exist_ok=True)
        for irrepr in young.Partition.partitions(N):
            save_matrices(output_path, irrepr, args.overwrite)
        print(f"Saved generators in [link=file:{output_path}]{output_path}[/link]")
    elif args.partition is not None:
        irrepr = young.Partition(args.partition)
        N = sum(irrepr)
        if N < 2:
            print(f"Generators for [b]S{N}[/b] are not required.")
            return 1
        print(f"Constructing generators for representation {irrepr}")
        output_path /= f"S_{N}"
        output_path.mkdir(parents=True, exist_ok=True)
        save_matrices(output_path, irrepr, args.overwrite)
        print(f"Saved generators in [link=file:{output_path}]{output_path}[/link]")
    else:
        show_generators(output_path)
    return 0



import subprocess
from typing import Iterable
from typing import List

import autoeft.form
from autoeft import utils

script = """#-
Off statistics;
Format nospaces;
Dimension {sun_dimension};
AutoDeclare Index F;
{exprs}
contract 0;
.sort
#do i=1,{basis_dimension}
#write "%E", expr`i'
#enddo
.end
"""


def contract_basis(
    tensor: str,
    sbasis: Iterable[str],
    sun_dimension: int,
    basis_dimension: int,
) -> List[utils.Fraction]:
    exprs = "\n".join(
        f"Local expr{i} = ({tensor})*({element});"
        for i, element in enumerate(sbasis, start=1)
    )
    data = script.format(
        sun_dimension=sun_dimension,
        exprs=exprs,
        basis_dimension=basis_dimension,
    )
    command = [autoeft.form.executable, "-q", "-M", "-"]
    contract_cmd = subprocess.run(
        command,  # noqa: S603
        input=data,
        capture_output=True,
        text=True,
        check=True,
    )
    return [utils.Fraction(res) for res in contract_cmd.stdout.splitlines()]
